#include <libguile.h>
#include <guile/gh.h>
#include <glib.h>

#include "debug.h"
#include "math.h"
#include "mathutils.h"
#include "pool.h"

#include "cue.h"
#include "table.h"

#include "game.h"
#include "animation.h"
#include "gtulpas.h"
#include "poolview.h"
#include "poolfeedgl.h"
#include "guileglue.h"
#include "fyjik.h"
#include "variant.h"
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-util.h>

typedef struct {
  SCM		arg_list;
  SCM	       *event_type;
  SCM	       *info1;
  SCM	       *info2;
  SCM	       *state;
  SCM	       *game_state;
  SCM	       *answers;
} glue_ugly_anim_info;

#define ANIM_EVENT_ITEMS_NOF		5

glue_enum_item anim_event_items_sorted_by_value [ANIM_EVENT_ITEMS_NOF];
glue_enum_item anim_event_items_sorted_by_symbol [ANIM_EVENT_ITEMS_NOF] = {
 { SCM_UNDEFINED, "cue-ball-collision", S_Cue_Ball_Collision},
 { SCM_UNDEFINED, "ball-ball-collision", S_Ball_Ball_Collision},
 { SCM_UNDEFINED, "ball-pocketed", S_Ball_Pocketed},
 { SCM_UNDEFINED, "ball-jump-off", S_Ball_Jump_Off},
 { SCM_UNDEFINED, "ball-mantinel-collision", S_Ball_Mantinel_Collision},
};

glue_enum_info anim_event_info = {
  anim_event_items_sorted_by_symbol,
  anim_event_items_sorted_by_value,
  ANIM_EVENT_ITEMS_NOF
};

static GList	*skin_list = NULL;
Skin	        * default_ball_skin;
Skin	        * default_cue_skin;
Skin	        * default_mark_skin;
Skin	        * default_table_skin;
Skin	        * default_pocket_skin;
Skin	        * default_outline_skin;

static GList    *fyjik_skin_list = NULL;
FyjikSkin	*default_fyjik_skin;

static gpointer skin_find (GList *list, char *name, int length);
static Skin* glue_scm2skin (SCM s_skin, Skin *defskin, int pos, const char *subr);

#define SCM_STR_CMP(S,STR,L)  (SCM_LENGTH (S) == (L) && (strncmp (SCM_CHARS (S), STR, L) == 0))

#define SCM_DARRAYP(X,L) (SCM_NIMP(X) && ((SCM_TYP7(X) == scm_tc7_vector) ||		  \
			 (SCM_TYP7(X) == scm_tc7_wvect) || (SCM_TYP7(X) == scm_tc7_fvect) || \
			 (SCM_TYP7(X) == scm_tc7_dvect)) && (SCM_LENGTH(X) == (L)))

#define SCM_LARRAYP(X,L) (SCM_NIMP(X) && ((SCM_TYP7(X) == scm_tc7_vector) ||		  \
			 (SCM_TYP7(X) == scm_tc7_wvect) || (SCM_TYP7(X) == scm_tc7_ivect) || \
			 (SCM_TYP7(X) == scm_tc7_uvect)) && (SCM_LENGTH(X) == (L)))

PoolTable	  pool_table = { NULL, NULL, 0, NULL, NULL,  0, NULL, 0,  DEFAULT_TABLE_HEIGHT,  NULL, 1, 1};
static int	  current_nof_holes = -1,
		  current_nof_vertices = -1;
Game*	          inside_set_rules = NULL;

static char		 *s_error_not_now = "you can not call this function now";
static char		 *s_error_not_in_anim = "you can not call this function during animation";

static FyjikSkin
builtin_fyjik_skin = { 
  NULL,
  POOL_SPHERE_RADIUS, 1 , 

  0.955,    /* coef nepruzneho narazu kouli */                                       
  0.05,	    /* treni o podlozku */                                                   
  0.05,     /* coef treni dvou kouli pro predani rychlost -> uhlova rychlost */                 
  0.05,	    /* coef treni kouli pro predani uhlove rychlosti uhl r -> uhl r. :) */            
  0.85,	    /* koef nepruzneho narazu o hranu*/                                      
  0.005,    /* ubytek tychlosti kolem osy z */                                      	     
  0.014,    /* valivy odpor */                                                      	     
  0.2,	    /* ubytek uhlove rychlosti pri odrazu o hranu */                                                    
  0.2,	    /* nepruzny odraz koule od podlozky stolu */                            
  981	    /* gravity acceleration */                                              
};

SCM_PROC (s_glue_cue_new, "cue-new", 4,0,0, sglue_cue_new);

static SCM
sglue_cue_new (SCM s_name, SCM s_length, SCM s_top_width, SCM s_bottom_width)
{
  PoolCueTypeInfo	*cue;

  SCM_ASSERT (SCM_NIMP (s_name) && SCM_STRINGP (s_name), s_name, SCM_ARG1, s_glue_cue_new);
  SCM_ASSERT (scm_number_p (s_length) == SCM_BOOL_T, s_length, SCM_ARG2, s_glue_cue_new);
  SCM_ASSERT (scm_number_p (s_top_width) == SCM_BOOL_T, s_top_width, SCM_ARG3, s_glue_cue_new);
  SCM_ASSERT (scm_number_p (s_bottom_width) == SCM_BOOL_T, s_bottom_width, SCM_ARG4, s_glue_cue_new);

  SCM_DEFER_INTS;
  
  cue = g_new (PoolCueTypeInfo, 1);
  cue -> name = gh_scm2newstr (s_name, NULL);
  cue -> length = gh_scm2double (s_length);
  cue -> top_width = gh_scm2double (s_top_width);
  cue -> bottom_width = gh_scm2double (s_bottom_width);

  pool_cue_type_list = g_list_prepend (pool_cue_type_list, cue);
  if (pool_cue.cue_type == NULL)
    {
      pool_cue.cue_type = cue;
      pool_cue.rebuild_gl_list = 1;
    }

  SCM_ALLOW_INTS;

  return SCM_UNSPECIFIED;
}

SCM_PROC (s_glue_ball_in_hand, "ball-in-hand", 1,3,0, sglue_ball_in_hand);
static SCM
sglue_ball_in_hand (SCM s_id, SCM s_prompt, SCM s_validate, SCM s_after)
{
  BallInHand	*bih;
  int		 id;

  SCM_ASSERT (SCM_INUMP (s_id), s_id, SCM_ARG1, s_glue_ball_in_hand);

  id = SCM_INUM (s_id);
  SCM_ASSERT ((0 <= id) || (id < gtulpas -> game -> balls_nof), s_id,
	      SCM_ARG1, s_glue_ball_in_hand);
  SCM_ASSERT (s_prompt == SCM_UNDEFINED || s_prompt == SCM_BOOL_F ||
	      (SCM_NIMP (s_prompt) && SCM_STRINGP (s_prompt)),
	      s_prompt, SCM_ARG2, s_glue_ball_in_hand);
  SCM_ASSERT (s_validate == SCM_BOOL_T || s_validate == SCM_UNDEFINED ||
	      scm_procedure_p (s_validate) == SCM_BOOL_T, s_validate, 
	      SCM_ARG3, s_glue_ball_in_hand);
  SCM_ASSERT (s_after == SCM_UNDEFINED || scm_procedure_p (s_after) == SCM_BOOL_T,
	      s_after, SCM_ARG4, s_glue_ball_in_hand);

  bih = ball_in_hand_new (id);

  bih -> validate = scm_procedure_p (s_validate) == SCM_BOOL_T ? s_validate : SCM_BOOL_T;
  bih -> after = scm_procedure_p (s_after) == SCM_BOOL_T ? s_after : SCM_BOOL_T;

  if (SCM_NIMP (s_validate))
    scm_protect_object (s_validate);

  if (SCM_NIMP (s_after))
    scm_protect_object (s_after);

  if (SCM_NIMP (s_prompt) && SCM_STRINGP (s_prompt))
    {
      g_free (bih -> prompt);
      bih->prompt = g_strndup (SCM_CHARS (s_prompt),
			       SCM_LENGTH (s_prompt));
    }

  game_add_event_nop_type_data (gtulpas->game, BALL_IN_HAND, (gpointer) bih);
  return SCM_UNSPECIFIED;
}

SCM_PROC (s_glue_ball_can_move, "ball-can-move", 1, 1, 0, sglue_ball_can_move);

static SCM
sglue_ball_can_move (SCM s_id, SCM s_verify)
{
  int	id;
  Game	*g;

  SCM_ASSERT (s_id == SCM_BOOL_F || SCM_INUMP (s_id), s_id, SCM_ARG1, s_glue_ball_can_move);
  SCM_ASSERT (s_verify == SCM_BOOL_T || s_verify == SCM_UNDEFINED ||
	      scm_procedure_p (s_verify) == SCM_BOOL_T,
	      s_verify, SCM_ARG2, s_glue_ball_can_move);

  g = gtulpas -> game;
  if (SCM_NIMP (g->ball_move_verify_func))
    scm_unprotect_object (g->ball_move_verify_func);

  if (s_id == SCM_BOOL_F)
    {
      g -> ball_move_id = -1;
      g -> ball_move_verify_func = SCM_BOOL_F;
    }
  else
    {
      id = SCM_INUM (s_id);


      SCM_ASSERT ((0 <= id) && (id < g -> balls_nof), s_id, SCM_ARG1, s_glue_ball_can_move);
      
      g -> ball_move_id = id;
      g -> ball_move_verify_func = (scm_procedure_p (s_verify) == SCM_BOOL_T) ? s_verify : SCM_BOOL_T;

      if (SCM_NIMP (s_verify))
	scm_protect_object (s_verify);
    }

  return SCM_UNDEFINED;
}

#define INSIDE_TABLE_2(DX, DY)		\
  tmp.p [X] = binfo -> p [X] + DX * radius;	\
  tmp.p [Y] = binfo -> p [Y] + DY * radius;	\
  if (ball_out_of_table (&tmp, -1) != 0)	\
    return FALSE

#define INSIDE_TABLE_1(ALFA)	\
  INSIDE_TABLE_2 (sin(ALFA), cos (ALFA))
  
int pool_place_ball_is_correct (int id, BallInHand *bih_info,
				PoolBallSpaceInfo *binfo, PoolBallSpaceInfo *frame)
{
  int			i;
  PoolBallSpaceInfo	tmp;
  dbl			radius;

  radius = binfo -> ball -> ball_type -> radius;

  tmp.ball = binfo -> ball;
  tmp.p [Z] = radius;

  INSIDE_TABLE_2 (0,0);
  if (ball_in_hole (binfo))
    return FALSE;

  INSIDE_TABLE_2 ( 0, 1);
  INSIDE_TABLE_2 ( 1, 0);
  INSIDE_TABLE_2 ( 0,-1);
  INSIDE_TABLE_2 (-1, 0);

  INSIDE_TABLE_1 (PI/4.);
  INSIDE_TABLE_1 (PI/4. + PI/2.);
  INSIDE_TABLE_1 (PI/4. + PI);
  INSIDE_TABLE_1 (PI/4. - PI/2.);

  for (i = 0; i < gtulpas -> game -> balls_nof; i++)
    if (i != id)
      if (euclid (binfo->p, frame[i].p) <
	  radius + frame[i].ball->ball_type->radius + EPSILON)
	return FALSE;

  if (bih_info && bih_info -> validate != SCM_BOOL_T &&
      glue_safe_call2 (bih_info -> validate, 
		       gh_double2scm (binfo -> p [X]),
		       gh_double2scm (binfo -> p [Y])) == SCM_BOOL_F)
    return FALSE;

  return TRUE; 
}

#undef INSIDE_TABLE_1
#undef INSIDE_TABLE_2


SCM_PROC(s_glue_place_ball, "place-ball", 2, 1, 0, sglue_place_ball);

static SCM
sglue_place_ball (SCM s_id, SCM s_position, SCM s_angles)
{
  PoolBallSpaceInfo	tmp, *b, *start;
  int			id;
  
  if (! animation_finished_p ())
    scm_misc_error (s_glue_place_ball, s_error_not_in_anim, SCM_EOL);

  SCM_ASSERT (SCM_INUMP(s_id), s_id, SCM_ARG1, s_glue_place_ball);
  SCM_ASSERT (SCM_DARRAYP(s_position,2), s_position, SCM_ARG2, s_glue_place_ball);
  SCM_ASSERT (s_angles == SCM_UNDEFINED || SCM_DARRAYP (s_angles, 2), s_angles, SCM_ARG3, s_glue_place_ball);

  id = SCM_INUM (s_id);
  SCM_ASSERT ((0 <= id) && (id < gtulpas->game->animation->nof_balls), s_id, SCM_ARG1, s_glue_place_ball);


  start = ANIMATION_LAST_FRAME (gtulpas->game->animation);
  b = start + id; tmp = *b; gh_scm2doubles (s_position, tmp.p); 

  /* put ball on table */
  tmp.p[Z] = b->ball->ball_type->radius;
  

  /* check whether position is free */
  if (pool_place_ball_is_correct (id, NULL, &tmp, start) == FALSE)
    return SCM_BOOL_F;

  *b = tmp;

  if (s_angles == SCM_UNDEFINED)
    quat_load_identity (b->q);
  else
    {
      dbl	angles [2];

      gh_scm2doubles (s_angles, angles);
      quat_load_from_angles (b->q,  DEG2RAD (angles [0]),  DEG2RAD (angles [1]));
    }  

  pool_view_redraw_mark_all ();
  return SCM_BOOL_T;
}


SCM_PROC (s_glue_place_cue, "place-cue", 1, 1, 0, sglue_place_cue);

static SCM
sglue_place_cue (SCM s_ball, SCM s_orientation)
{
  int			id;
  dbl			orientation [2];

  if (! animation_finished_p ())
    scm_misc_error (s_glue_place_cue, s_error_not_in_anim, SCM_EOL);

  SCM_ASSERT (SCM_INUMP(s_ball), s_ball, SCM_ARG1, s_glue_place_cue);
  SCM_ASSERT ((s_orientation == SCM_UNDEFINED) || SCM_DARRAYP (s_orientation, 2), 
	      s_orientation, SCM_ARG2, s_glue_place_cue);

  id = SCM_INUM (s_ball);

  SCM_ASSERT ((0 <= id) && (id < gtulpas->game->animation->nof_balls), s_ball, 
	      SCM_ARG1, s_glue_place_cue);

  cue_set_ball (&pool_cue, id);
  if (s_orientation != SCM_UNDEFINED)
    {
      gh_scm2doubles (s_orientation, orientation);
      cue_set_orientation (&pool_cue, 
			   DEG2RAD (orientation[0]), 
			   DEG2RAD (orientation [1]));
    }

  cue_recalc_hints (NULL);
  pool_view_redraw_mark_all ();

  return SCM_UNSPECIFIED;
}

/*
 *
 *		TABLE
 */

SCM_PROC (s_glue_table_start, "table-start", 2, 0, 0, sglue_table_start);

static SCM
sglue_table_start (SCM s_nof_vertices, SCM s_nof_holes)
{
  int		nof_vertices;
  int		nof_holes;

  if (current_nof_vertices >= 0 || (! inside_set_rules))
     scm_misc_error (s_glue_table_start, s_error_not_now, SCM_EOL);

  SCM_ASSERT (SCM_INUMP (s_nof_vertices), s_nof_vertices, SCM_ARG1, s_glue_table_start);
  SCM_ASSERT (SCM_INUMP (s_nof_holes), s_nof_holes, SCM_ARG2, s_glue_table_start);

  nof_vertices = SCM_INUM (s_nof_vertices);
  nof_holes = SCM_INUM (s_nof_holes);

  SCM_ASSERT (nof_vertices > 0, nof_vertices, SCM_ARG1, s_glue_table_start);
  SCM_ASSERT (nof_holes >= 0, nof_vertices, SCM_ARG1, s_glue_table_start);

  if (pool_table.vertices) g_free (pool_table.vertices);
  if (pool_table.holes) g_free (pool_table.vertices);
  if (pool_table.outer_vertices) g_free (pool_table.outer_vertices);
  pool_table.outer_vertices = NULL;

  pool_table.table_height = DEFAULT_TABLE_HEIGHT;
  pool_table.skin = default_table_skin;
  pool_table.nof_vertices = nof_vertices;
  pool_table.vertices = g_new (dbl, 2 * nof_vertices);
  
  pool_table.nof_holes = nof_holes;
  pool_table.holes = (nof_holes == 0) ? NULL : g_new (TableHole, nof_holes);

  pool_table.update_view_info = 1;
  pool_table.update_fyjik_info = 1;

  current_nof_vertices = 0;
  current_nof_holes = 0;

  return SCM_UNSPECIFIED;
}

SCM_PROC (s_glue_table_end, "table-end", 0, 0, 0, sglue_table_end);

static SCM
sglue_table_end (void)
{
  if (current_nof_holes != pool_table.nof_holes)
    scm_misc_error (s_glue_table_end, "more pockets declared than defined", SCM_EOL); 

  if (current_nof_vertices != pool_table.nof_vertices)
    scm_misc_error (s_glue_table_end, "more vertices declared than defined", SCM_EOL); 

  current_nof_vertices = -1;
  current_nof_vertices = -1;

  return SCM_UNSPECIFIED;
}

SCM_PROC (s_glue_vertex, "vertex", 1, 0, 0, sglue_vertex);

static SCM
sglue_vertex (SCM s_coord)
{
  if (current_nof_vertices < 0)
    scm_misc_error (s_glue_vertex, s_error_not_now, SCM_EOL);

  if (current_nof_vertices == pool_table.nof_vertices)
    scm_misc_error (s_glue_vertex, "too many vertices", SCM_EOL);

  SCM_ASSERT (SCM_DARRAYP (s_coord, 2), s_coord, SCM_ARG1, s_glue_vertex);

  gh_scm2doubles (s_coord, pool_table.vertices + (2 * current_nof_vertices));
  current_nof_vertices ++;

  return SCM_UNSPECIFIED ;
}

SCM_PROC (s_glue_hole, "pocket", 2, 1, 0, sglue_hole);

static SCM
sglue_hole (SCM s_coord, SCM s_radius, SCM s_skin)
{
  TableHole	*hole;
  Skin		*skin;

  if (current_nof_holes < 0)
    scm_misc_error (s_glue_hole, s_error_not_now, SCM_EOL);

  if (current_nof_holes == pool_table.nof_holes)
    scm_misc_error (s_glue_hole, "too many pockets", SCM_EOL);

  SCM_ASSERT (SCM_DARRAYP (s_coord, 2), s_coord, SCM_ARG1, s_glue_hole);
  SCM_ASSERT (scm_number_p (s_radius) == SCM_BOOL_T, s_radius, SCM_ARG2, s_glue_hole);
  skin = glue_scm2skin (s_skin, default_pocket_skin, SCM_ARG3, s_glue_hole);

  hole = pool_table.holes + current_nof_holes;
  hole -> hid = current_nof_holes++;
  hole -> skin = skin;

  gh_scm2doubles (s_coord, hole->p);
  hole->radius = gh_scm2double (s_radius);

  SCM_ASSERT (hole->radius > 0, s_radius, SCM_ARG2, s_glue_hole);

  return SCM_UNSPECIFIED;
}

SCM_PROC (s_glue_table_outline, "table-outline", 1,0,1, sglue_table_outline);

static SCM
sglue_table_outline (SCM s_skin, SCM lst)
{
  Skin	*skin;
  SCM	data;
  int	length, i;
  dbl  *ptr;

  if (current_nof_holes < 0)
    scm_misc_error (s_glue_table_outline, s_error_not_now, SCM_EOL);

  skin = glue_scm2skin (s_skin, NULL, SCM_ARG1, s_glue_table_outline);

  length = scm_ilength (lst);
  SCM_ASSERT (length > 0, lst, SCM_ARG2, s_glue_table_outline);

  if (pool_table.outer_vertices != NULL)
    g_free (pool_table.outer_vertices);

  pool_table.nof_outer_vertices = length;
  ptr = pool_table.outer_vertices = g_new (dbl, 2 * length);

  for (i = 0; i < length; i++, ptr += 2)
    {
      data = SCM_CAR (lst);
      lst = SCM_CDR (lst);

      SCM_ASSERT (SCM_DARRAYP (data,2), data, (i < 6) ? (SCM_ARG2 + i) : SCM_ARGn, s_glue_table_outline);
      gh_scm2doubles (data, ptr);
    }

  return SCM_UNSPECIFIED;
}

SCM_PROC (s_glue_cushion_height, "cushion-height", 1,0,0, sglue_cushion_height);
static SCM
sglue_cushion_height (SCM s_height)
{
  dbl height = -1;

  if (scm_number_p (s_height) == SCM_BOOL_T)
    height = gh_scm2double (s_height);

  SCM_ASSERT (height > 0., s_height, SCM_ARG1, s_glue_cushion_height);
  pool_table.table_height = height;
  pool_table.update_fyjik_info = 1;
  pool_table.update_view_info = 1;

  return SCM_UNDEFINED;
}

static Skin*
glue_scm2skin (SCM s_skin, Skin *defskin, int pos, const char *subr)
{
  GList	*lskin;

  if (s_skin == SCM_UNDEFINED && defskin)
    return defskin;

  SCM_ASSERT (SCM_NIMP (s_skin) && SCM_ROSTRINGP (s_skin), s_skin, pos, subr);
  lskin = skin_find (skin_list, SCM_CHARS (s_skin), SCM_LENGTH (s_skin));

  if (lskin == NULL)
    scm_misc_error (subr, "No such skin %S", scm_cons (s_skin, SCM_EOL));

  return lskin->data;
}

SCM_PROC (s_glue_mark_point, "mark-point", 2, 1, 0, sglue_mark_point);

static SCM
sglue_mark_point (SCM s_coord, SCM s_size, SCM s_skin)
{
  TableMark	*mark;
  Skin		*skin;

  SCM_ASSERT (SCM_DARRAYP (s_coord, 2), s_coord, SCM_ARG1, s_glue_mark_point);
  SCM_ASSERT (scm_number_p (s_size) == SCM_BOOL_T , s_size, SCM_ARG2, s_glue_mark_point);
  skin = glue_scm2skin (s_skin, default_mark_skin, SCM_ARG3, s_glue_mark_point);

  if (current_nof_holes < 0)
    scm_misc_error (s_glue_hole, s_error_not_now, SCM_EOL);

  mark = g_new (TableMark, 1);
  pool_table.mark_list = g_list_prepend (pool_table.mark_list, mark);

  mark -> mark_type = TableMark_Point;
  mark -> skin = skin;

  gh_scm2doubles (s_coord, mark->p1);
  mark -> dparam = gh_scm2double (s_size);
  mark -> data = NULL;

  return SCM_UNSPECIFIED ;
}

SCM_PROC (s_glue_mark_line, "mark-line", 2, 1, 0, sglue_mark_line);

static SCM
sglue_mark_line (SCM s_coord1, SCM s_coord2, SCM s_skin)
{
  TableMark	*mark;
  Skin		*skin;

  if (current_nof_holes < 0)
    scm_misc_error (s_glue_hole, s_error_not_now, SCM_EOL);

  SCM_ASSERT (SCM_DARRAYP (s_coord1, 2), s_coord1, SCM_ARG1, s_glue_mark_line);
  SCM_ASSERT (SCM_DARRAYP (s_coord2, 2), s_coord2, SCM_ARG2, s_glue_mark_line);
  skin = glue_scm2skin (s_skin, default_mark_skin, SCM_ARG3, s_glue_mark_point);

  mark = g_new (TableMark, 1);
  pool_table.mark_list = g_list_prepend (pool_table.mark_list, mark);

  mark -> mark_type = TableMark_Line;
  mark -> skin = skin;
  mark -> data = NULL;

  gh_scm2doubles (s_coord1, mark->p1);
  gh_scm2doubles (s_coord2, mark->p2);

  return SCM_UNSPECIFIED ;
}

SCM_PROC (s_glue_mark_diamond, "mark-diamond", 3, 1, 0, sglue_mark_diamond);

static SCM
sglue_mark_diamond (SCM s_center, SCM x_size, SCM y_size, SCM s_skin)
{
  TableMark	*mark;
  Skin		*skin;

  if (current_nof_holes < 0)
    scm_misc_error (s_glue_hole, s_error_not_now, SCM_EOL);

  SCM_ASSERT (SCM_DARRAYP (s_center, 2), s_center, SCM_ARG1, s_glue_mark_diamond);
  SCM_ASSERT (scm_number_p (x_size), x_size, SCM_ARG2, s_glue_mark_diamond);
  SCM_ASSERT (scm_number_p (y_size), y_size, SCM_ARG3, s_glue_mark_diamond);
  skin = glue_scm2skin (s_skin, default_mark_skin, SCM_ARG4, s_glue_mark_diamond);

  mark = g_new (TableMark,1);
  pool_table.mark_list = g_list_prepend (pool_table.mark_list, mark);

  mark -> mark_type = TableMark_Diamond;
  mark -> skin = skin;
  mark -> data = NULL;

  gh_scm2doubles (s_center, mark->p1);
  mark -> p2 [X] = gh_scm2double (x_size);
  mark -> p2 [Y] = gh_scm2double (y_size);

  return SCM_UNSPECIFIED;
}

SCM_PROC (s_glue_mark_arc, "mark-arc", 2, 4, 0, sglue_mark_arc);

static SCM
sglue_mark_arc (SCM s_center, SCM s_radius, SCM s_precision, SCM s_start, SCM s_end, SCM s_skin)
{
  TableMark	*mark;
  Skin		*skin;

  if (current_nof_holes < 0)
    scm_misc_error (s_glue_hole, s_error_not_now, SCM_EOL);

  SCM_ASSERT (SCM_DARRAYP (s_center, 2), s_center, SCM_ARG1, s_glue_mark_arc);
  SCM_ASSERT (scm_number_p (s_radius) == SCM_BOOL_T, s_radius, SCM_ARG2, s_glue_mark_arc);
  SCM_ASSERT (SCM_INUMP (s_precision) || s_precision == SCM_UNDEFINED,
	      s_precision, SCM_ARG3, s_glue_mark_arc);
  SCM_ASSERT ((s_start == SCM_UNDEFINED && s_end == SCM_UNDEFINED) ||
	      (scm_number_p (s_start) == SCM_BOOL_T && scm_number_p (s_end) == SCM_BOOL_T),
	      s_start, SCM_ARG4, s_glue_mark_arc);
  skin = glue_scm2skin (s_skin, default_mark_skin, SCM_ARG6, s_glue_mark_arc);

  mark = g_new (TableMark, 1);
  pool_table.mark_list = g_list_prepend (pool_table.mark_list, mark);

  mark -> mark_type = TableMark_Arc;
  mark -> skin = skin;
  mark -> data = NULL;

  gh_scm2doubles (s_center, mark->p1);
  mark -> dparam = gh_scm2double (s_radius);
  mark -> iparam = s_precision == SCM_UNDEFINED ? 8 : SCM_INUM (s_precision);

  if (s_start == SCM_UNDEFINED)
    { mark->p2 [0] = 0; mark->p2 [1] = PI * 2.; }
  else
    {
      mark -> p2 [0] = DEG2RAD (gh_scm2double (s_start));
      mark -> p2 [1] = DEG2RAD (gh_scm2double (s_end));
    }

  return SCM_UNSPECIFIED;
}

/*
 *
 *	Views....
 */

SCM_PROC (s_default_view, "default-view", 1,0,0, sdefault_view);

static SCM
sdefault_view (SCM s_id)
{
  int		 id = -1;

  if (! inside_set_rules)
    scm_misc_error (s_default_view, s_error_not_now, SCM_EOL);

  if (SCM_INUMP (s_id))
    id = SCM_INUM (s_id) - 1;

  SCM_ASSERT ((0 <= id) && (id < GAME_MAX_NOF_VIEW_POS) &&
	      (inside_set_rules -> view_positions [id] . eye_focus_distance >= 0),
	      s_id, SCM_ARG1, s_default_view);

  inside_set_rules -> default_view_position = id;

  return SCM_UNSPECIFIED;
}

SCM_PROC (s_glue_view, "view", 5, 0, 0, sglue_view);

static SCM
sglue_view (SCM s_id, SCM s_alfa, SCM s_beta, SCM s_distance, SCM s_center)
{
  ViewPosition	*view;
  int		 id = -1;

  if (! inside_set_rules)
    scm_misc_error (s_glue_view, s_error_not_now, SCM_EOL);


  if (SCM_INUMP (s_id))
    id = SCM_INUM (s_id) - 1;
  
  SCM_ASSERT ((0 <= id) && (id < GAME_MAX_NOF_VIEW_POS), s_id, SCM_ARG1, s_glue_view);
  SCM_ASSERT (scm_number_p (s_alfa) == SCM_BOOL_T, s_alfa, SCM_ARG2, s_glue_view);
  SCM_ASSERT (scm_number_p (s_beta) == SCM_BOOL_T, s_beta, SCM_ARG3, s_glue_view);
  SCM_ASSERT (scm_number_p (s_distance) == SCM_BOOL_T, s_distance, SCM_ARG4, s_glue_view);
  SCM_ASSERT (SCM_DARRAYP (s_center, 3), s_center, SCM_ARG5, s_glue_view);

  view = inside_set_rules -> view_positions + id;

  view -> alfa = DEG2RAD (gh_scm2double (s_alfa));
  view -> beta = DEG2RAD (gh_scm2double (s_beta));
  view -> eye_center_distance = gh_scm2double (s_distance);
  view -> eye_focus_distance = POOL_INIT_EYE_CENTER;

  gh_scm2doubles (s_center, view -> center_of_rotation);
  
  return SCM_UNSPECIFIED;
}

/*
 *
 *
 *	Skins...
 */

static struct {
  char		*ptr;
  int		length;
} skin_find_data;

static int
find_custom_skin_cb (Skin *skin)
{
  if ((int) strlen (skin->name) != skin_find_data.length)
    return 1; /* different */
  return strncmp (skin -> name, skin_find_data.ptr, skin_find_data.length);
}

static gpointer
skin_find (GList *list, char *name, int length)
{
  skin_find_data.ptr = name;
  skin_find_data.length = length;
  return g_list_find_custom (list, NULL, (GCompareFunc) find_custom_skin_cb);
}

SCM_PROC (s_glue_skin, "skin", 2, 3, 1, sglue_skin);
static SCM
sglue_skin (SCM s_name, SCM s_color1, SCM s_color2, SCM s_texfile, SCM s_dvalue, SCM s_rest)
{
  SCM		 data;
  GList		*lskin;
  Skin		*skin;
  int		 has_strip = 0,
		 has_mark = 0;

  SCM_ASSERT (SCM_NIMP (s_name) && (SCM_STRINGP (s_name) || SCM_SYMBOLP (s_name))
	      , s_name, SCM_ARG1, s_glue_skin);
  SCM_ASSERT (SCM_DARRAYP (s_color1, 3), s_color1, SCM_ARG2, s_glue_skin);
  SCM_ASSERT (s_color2 == SCM_UNDEFINED || SCM_DARRAYP (s_color2, 3),
	      s_color2, SCM_ARG3, s_glue_skin);
  SCM_ASSERT (s_texfile == SCM_BOOL_F || s_texfile == SCM_UNDEFINED ||
	      (SCM_NIMP (s_texfile) && SCM_STRINGP (s_texfile)),
	      s_texfile, SCM_ARG1, s_glue_skin);
  SCM_ASSERT (s_dvalue == SCM_UNDEFINED || scm_number_p (s_dvalue) == SCM_BOOL_T,
	      s_dvalue, SCM_ARG5, s_glue_skin);

  while (s_rest != SCM_EOL)
    {
      if (! SCM_CONSP (s_rest))
	break;

      data = SCM_CAR (s_rest);
      if (! (SCM_NIMP (data) && SCM_SYMBOLP (data)))
	break;

      if (SCM_STR_CMP (data, "has-mark", 8))
	has_mark = TRUE;
      else if (SCM_STR_CMP (data, "has-strip", 9))
	has_strip = TRUE;
      else
	break;

      s_rest = SCM_CDR (s_rest);
    }

  SCM_ASSERT (s_rest == SCM_EOL, s_rest, SCM_ARGn, s_glue_skin);
  lskin = skin_find (skin_list, SCM_CHARS (s_name), SCM_LENGTH (s_name));
  
  if (lskin)
    {
      skin = (Skin *) (lskin -> data);
      if (skin -> texture_file)
	free (skin -> texture_file);
      /* FIXME: remove also texture data... */
    }
  else
    {
      skin = g_new (Skin, 1);
      skin_list = g_list_prepend (skin_list, skin);
      skin -> name = gh_scm2newstr (s_name, NULL);

      skin -> color2 [0] = 0;
      skin -> color2 [1] = 0;
      skin -> color2 [2] = 0;
    }
  
  gh_scm2doubles (s_color1, skin -> color1);
  if (s_color2 != SCM_UNDEFINED)
    gh_scm2doubles (s_color2, skin -> color2);

  skin -> texture_file = ((s_texfile == SCM_BOOL_F) ||(s_texfile == SCM_UNDEFINED)) 
    ? NULL
    : gh_scm2newstr (s_texfile, NULL);
  skin -> dvalue = (s_dvalue == SCM_UNDEFINED) ? 0.0 : gh_scm2double (s_dvalue);

  skin -> has_mark = has_mark;
  skin -> has_strip = has_strip;

  return SCM_UNSPECIFIED;
}

SCM_PROC (s_glue_fyjik_skin, "fyjik-skin", 1, 0, 1, sglue_fyjik_skin);

static SCM
sglue_fyjik_skin (SCM s_name, SCM s_rest)
{
  SCM			 name, data;
  dbl			 radius, weight;
  int			 set_radius = 0,
			 set_weight = 0;
  FyjikSkin		*fskin;
  GList			*lskin;

  SCM_ASSERT (SCM_NIMP (s_name) && SCM_ROSTRINGP (s_name), s_name, SCM_ARG1, s_glue_fyjik_skin);

  while (s_rest != SCM_EOL)
    {
      if (! (SCM_CONSP (s_rest) && SCM_CONSP (SCM_CDR (s_rest))))
	break;

      name = SCM_CAR (s_rest);
      data = SCM_CADR (s_rest);

      if (! (SCM_NIMP (name) && SCM_SYMBOLP (name)))
	break;

      if (scm_number_p (data) == SCM_BOOL_F)
	break;

      if (SCM_STR_CMP (name, "radius", 6))
	{ set_radius = TRUE; radius = gh_scm2double (data); }
      else if (SCM_STR_CMP (name, "diameter", 8))
	{ set_radius = TRUE; radius = gh_scm2double (data) / 2.0; }
      else if (SCM_STR_CMP (name, "weight", 6))
	{ set_weight = TRUE; weight = gh_scm2double (data); }
      else break;

      s_rest = SCM_CDDR (s_rest);
    }

  SCM_ASSERT (s_rest == SCM_EOL, s_rest, SCM_ARGn, s_glue_fyjik_skin);
  
  lskin = skin_find (fyjik_skin_list, SCM_CHARS (s_name), SCM_LENGTH (s_name));

  if (lskin)
    fskin = (FyjikSkin *) (lskin -> data);
  else
    {
      fskin = g_new (FyjikSkin, 1);
      *fskin = builtin_fyjik_skin;
      fskin -> name = gh_scm2newstr (s_name, NULL);

      fyjik_skin_list = g_list_prepend (fyjik_skin_list, fskin);
    }

  if (set_radius)
    fskin -> radius = radius;

  if (set_weight)
    fskin -> weight = weight;

  return SCM_UNSPECIFIED;
}


SCM_PROC (s_cue_set_skin, "cue-set-skin", 1,0,0, scue_set_skin);

static SCM
scue_set_skin (SCM s_skin)
{
  pool_cue.cue_skin = glue_scm2skin (s_skin, NULL, SCM_ARG1, s_cue_set_skin);
  pool_cue.rebuild_gl_list = TRUE;

  return SCM_UNSPECIFIED;
}

SCM_PROC (s_ball_set_skin, "ball-set-skin", 3, 1, 0, sball_set_skin);

static SCM
sball_set_skin (SCM s_id, SCM s_name, SCM s_skin, SCM s_fyjik_skin)
{
  GList	 *lfskin;
  Skin   *skin;
  FyjikSkin *fs;
  int	id = -1;

  if (! inside_set_rules)
     scm_misc_error (s_ball_set_skin, s_error_not_now, SCM_EOL);

  if (SCM_INUMP (s_id))
    id = SCM_INUM (s_id);

  SCM_ASSERT ((0 <= id) && (id < inside_set_rules -> balls_nof),
	      s_id, SCM_ARG1, s_ball_set_skin);
  SCM_ASSERT (SCM_NIMP (s_name) && SCM_STRINGP (s_name), s_name,
	      SCM_ARG2, s_ball_set_skin);
  skin = glue_scm2skin (s_skin, NULL, SCM_ARG3, s_ball_set_skin);
  SCM_ASSERT (s_fyjik_skin == SCM_UNDEFINED || 
	      (SCM_NIMP (s_fyjik_skin) && SCM_ROSTRINGP (s_fyjik_skin)),
	      s_fyjik_skin, SCM_ARG4, s_ball_set_skin);

  if (s_fyjik_skin == SCM_UNDEFINED)
    fs = default_fyjik_skin;
  else
    {
      lfskin = skin_find (fyjik_skin_list, SCM_CHARS (s_fyjik_skin), SCM_LENGTH (s_fyjik_skin));
      
      if (lfskin == NULL)
	scm_misc_error (s_ball_set_skin, "No such fyjik skin %S", scm_cons (s_fyjik_skin, SCM_EOL));

      fs = (FyjikSkin *) (lfskin->data);
    }

  inside_set_rules -> balls [id] . ball_skin = skin;
  inside_set_rules -> balls [id] . ball_type = fs;

  g_free (inside_set_rules -> balls [id] . name);
  inside_set_rules -> balls [id] . name = g_strndup (SCM_CHARS (s_name),
						    SCM_LENGTH (s_name));

  return SCM_UNSPECIFIED;
}


/*
 *
 *	Messages ...
 */

SCM_PROC (s_glue_message, "message", 1, 0, 0, sglue_message);

static SCM
sglue_message (SCM s_message)
{
  char		*message;

  SCM_ASSERT (SCM_NIMP (s_message) && SCM_STRINGP (s_message), s_message, SCM_ARG1, s_glue_message);

  message = gh_scm2newstr (s_message, NULL);
  gtulpas_status_message (message);
  scm_must_free (message);

  return SCM_UNSPECIFIED;
}

SCM_PROC (s_glue_message_start, "message-start", 1, 0, 0, sglue_message_start);

static SCM
sglue_message_start (SCM s_message)
{
  char		*message;

  SCM_ASSERT (SCM_NIMP (s_message) && SCM_STRINGP (s_message), s_message, SCM_ARG1, s_glue_message_start);

  message = gh_scm2newstr (s_message, NULL);
  gtulpas_status_start (message);
  scm_must_free (message);

  return SCM_UNSPECIFIED;
}

SCM_PROC (s_glue_message_end, "message-end", 0, 0, 0, sglue_message_end);
static SCM
sglue_message_end (void)
{
  gtulpas_status_end ();
  return SCM_UNSPECIFIED;
}

typedef enum {
  GluePRTableSetupFunc,
  GluePRNewGameFunc,
  GluePREventHandlerFunc,
  GluePRNextShotFunc,
  GluePREndOfShotFunc,
  GluePRNofBalls,
  GluePRNofPlayers,
  GluePRHelpFile,
  GluePREventMask,
  GluePRQueries,
  GluePRScore,
  GluePROptions,
  GluePRGameStateDeepCopy,
  GluePRSafeToUndo,
  __GluePRCount
} glue_pool_rules_params;

#define GLUE_POOL_RULES_NOF_PARAMS		__GluePRCount

static glue_enum_item glue_pool_rules_params_sby_value [GLUE_POOL_RULES_NOF_PARAMS];
static glue_enum_item glue_pool_rules_params_sby_symbol [GLUE_POOL_RULES_NOF_PARAMS] = {
  { SCM_UNDEFINED, "table-setup", GluePRTableSetupFunc},
  { SCM_UNDEFINED, "new-game", GluePRNewGameFunc},
  { SCM_UNDEFINED, "event-handler", GluePREventHandlerFunc},
  { SCM_UNDEFINED, "next-shot", GluePRNextShotFunc},
  { SCM_UNDEFINED, "end-of-shot", GluePREndOfShotFunc},
  { SCM_UNDEFINED, "number-of-balls", GluePRNofBalls},
  { SCM_UNDEFINED, "number-of-players", GluePRNofPlayers},
  { SCM_UNDEFINED, "help-file", GluePRHelpFile},
  { SCM_UNDEFINED, "event-mask", GluePREventMask},
  { SCM_UNDEFINED, "queries", GluePRQueries},
  { SCM_UNDEFINED, "score", GluePRScore},
  { SCM_UNDEFINED, "options", GluePROptions},
  { SCM_UNDEFINED, "game-state-deep-copy", GluePRGameStateDeepCopy},
  { SCM_UNDEFINED, "safe-to-undo", GluePRSafeToUndo}
};

static glue_enum_info glue_pool_rules_info = {
  glue_pool_rules_params_sby_symbol,
  glue_pool_rules_params_sby_value,
  GLUE_POOL_RULES_NOF_PARAMS
};

static glue_enum_item glue_score_types_sby_value [SVBalls+1];
static glue_enum_item glue_score_types_sby_symbol [SVBalls+1] = {
  { SCM_UNDEFINED, "counter", SNumbers},
  { SCM_UNDEFINED, "ball-counter", SBalls},
  { SCM_UNDEFINED, "faul-counter", SFauls},
  { SCM_UNDEFINED, "balls", SVBalls}
};

static glue_enum_info glue_score_types_info = {
  glue_score_types_sby_symbol,
  glue_score_types_sby_value,
  SVBalls+1
};

#define SET_PROCEDURE(NAME)				\
	if (scm_procedure_p (data) == SCM_BOOL_F)	\
	  goto error;					\
							\
	if (SCM_NIMP (game -> NAME))			\
	  scm_unprotect_object (game->NAME);		\
							\
        scm_protect_object (data);			\
	game -> NAME = data

#define SET_PROCEDURE_ALLOW_F(NAME)			\
	if (data != SCM_BOOL_F)				\
	  {						\
	    if (scm_procedure_p (data) == SCM_BOOL_F)	\
	      goto error;				\
	    else					\
	      scm_protect_object (data);		\
	  }						\
							\
	if (SCM_NIMP (game -> NAME))			\
	  scm_unprotect_object (game -> NAME);		\
							\
	game -> NAME = data


static int
glue_game_compare_cb (Game *game, char *name)
{
  return strcmp (game -> name, name);
}

static int 
glue_parse_event_mask (Game *game, SCM data, int pos, const char *name)
{
  AnimEventType	ae;
  int			i;
  
  if (! (scm_ilength (data) >= 0))
    goto error;

  game -> anim_event_mask = 0;
  while (data != SCM_EOL)
    {
      ae = glue_enum_scm2value (SCM_CAR (data), &anim_event_info);
      data = SCM_CDR (data);

      i = -1;
      if (SCM_NIMP (data) && SCM_CONSP (data) && SCM_INUMP (SCM_CAR(data)))
	{ 
	  i = SCM_INUM (SCM_CAR(data));
	  data = SCM_CDR (data);
	  pos++;
	  if (i <= 0)
	    goto error;
	}

      switch (ae)
	{
	case S_Ball_Mantinel_Collision:
	  game -> anim_event_mask |= ae;
	  game -> anim_event_ball_mantinel_num = i;
	  break;

	case S_Ball_Ball_Collision:
	  game -> anim_event_mask |= ae;
	  game -> anim_event_ball_ball_num = i;
	  break;

	case S_Cue_Ball_Collision:
	case S_Ball_Pocketed:
	case S_Ball_Jump_Off:
	  game -> anim_event_mask |= ae;
	  break;

	case -1:
	default:
	  goto error;
	}
      pos++;
    }

  return 0;

 error:
  if (name)
    SCM_ASSERT (0, data, (pos <= SCM_ARG7) ? pos : SCM_ARGn, name);

  return -1;
}

SCM_PROC(s_glue_pool_rules, "pool-rules", 1, 0, 1, sglue_pool_rules);

static SCM
sglue_pool_rules (SCM s_name, SCM s_rest)
{
  GList		*game_found;
  SCM		data;
  SCM		nval;
  Game		*game;
  char		*name;
  int		 pid;

  SCM_ASSERT (SCM_NIMP (s_name) && SCM_STRINGP(s_name), s_name, SCM_ARG1, s_glue_pool_rules);

  name = GLUE_STRDUP (s_name);

  game_found = g_list_find_custom (game_list, name, (GCompareFunc) glue_game_compare_cb);
  if (game_found == NULL)
    {
      game = game_new (2,2,3);
      game -> name = name;
      game->id = game_max_id++;
    }
  else
    {
      game = (Game *) (game_found -> data);
      g_free (name);
    }

  while (s_rest != SCM_EOL)
    {
      if (! SCM_CONSP (SCM_CDR (s_rest)))
	goto error;

      pid = glue_enum_scm2value (SCM_CAR (s_rest), &glue_pool_rules_info);
      if (pid == -1)
	goto error;

      data = SCM_CADR (s_rest);
      nval = SCM_CDDR (s_rest);
      
      switch ((glue_pool_rules_params) pid)
	{
	case GluePRTableSetupFunc:
	  SET_PROCEDURE (setup_table_func);
	  break;

	case GluePRNewGameFunc:
	  SET_PROCEDURE (new_game_func);
	  break;

	case GluePREventHandlerFunc:
	  SET_PROCEDURE (game_rules_func);
	  break;

	case GluePRNextShotFunc:
	  SET_PROCEDURE_ALLOW_F (next_stroke_func);
	  break;

	case GluePREndOfShotFunc:
	  SET_PROCEDURE_ALLOW_F (end_of_shot_func);
	  break;

	case GluePRNofBalls:
	  {
	    int		i;

	    if (! SCM_INUMP (data))
	      goto error;

	    i = SCM_INUM (data);
	    if (i != game -> balls_nof)
	      game_set_nof_balls (game, i);

	    break;
	  }
	
	case GluePRNofPlayers:
	  {
	    long   newnof [2];

	    if (SCM_INUMP (data))
	      newnof [0] = newnof [1] = SCM_INUM (data);
	    else if (SCM_LARRAYP (data, 2))
	      gh_scm2longs (data, newnof);
	    else if (scm_list_p (data) == SCM_BOOL_T &&
		     gh_length (data) == 2)
	      {
		newnof [0] = gh_scm2long (SCM_CAR (data));
		newnof [1] = gh_scm2long (SCM_CADR (data));
	      }
	    else
	      goto error;

	    if (game -> max_players_nof < newnof [1])
	      {
		g_free (game -> player);
		game -> player = (Player **) g_new (Player *, newnof [1]);
	      }

	    game -> min_players_nof = newnof [0];
	    game -> max_players_nof = newnof [1];

	    break;
	  }

	case GluePRHelpFile:
	  if (! (SCM_NIMP (data) && SCM_STRINGP (data)))
	    goto error;

	  if (game -> help_file)
	    free (game -> help_file);

	  game -> help_file = GLUE_STRDUP (data);
	  break;

	case GluePREventMask:
	  if (glue_parse_event_mask (game, data, SCM_ARGn, NULL))
	    goto error;

	  break;

	case GluePRQueries:
	  query_free (game -> query_list);
	  game -> query_list = query_scm2list (data, SCM_ARGn, s_glue_pool_rules);
	  break;

	case GluePRScore:
	  {
	    SCM  par, type, name;
	    int  score;

	    if (scm_list_p (data) == SCM_BOOL_F)
	      goto error;

	    while (data != SCM_EOL)
	      {
		par = SCM_CAR (data);

		if (scm_ilength (par) != 2)
		  goto error;

		type = SCM_CAR (par);
		name = SCM_CADR (par);

		if (! (SCM_NIMP (name) && SCM_STRINGP (name)))
		  goto error;

		score = glue_enum_scm2value (type, &glue_score_types_info);
		if (score == -1)
		  goto error;

		SCM_COERCE_SUBSTR (name);
		game_add_score (game, score, SCM_CHARS (name));

		data = SCM_CDR (data);
	      }
	    break;
	  }
	case GluePROptions:
	  {
	    SCM		e = SCM_UNSPECIFIED;
	    
	    game -> variant_list = variant_scm2vlist (data, &e);
	    if (game -> variant_list == NULL && e != SCM_UNSPECIFIED)
	      { s_rest = e; goto error; }

	    break;
	  }

	case GluePRGameStateDeepCopy:
	  SET_PROCEDURE_ALLOW_F (game_state_deep_copy_func);
	  break;


	case GluePRSafeToUndo:
	  if (data == SCM_BOOL_F)
	    game -> safe_to_undo = FALSE;
	  else 
	    game -> safe_to_undo = TRUE;
	  break;

	default:
	  goto error;
	}

      s_rest = nval;
    }

  if (game->setup_table_func == SCM_BOOL_F || game->new_game_func == SCM_BOOL_F
      || game -> game_rules_func == SCM_BOOL_F || game->nscore == 0)
    scm_misc_error  (s_glue_pool_rules, 
		     "table-setup, new-game, event-handler and score have to be defined for %S game",
		     scm_cons (s_name, SCM_EOL));

  if (! game_found)
    {
      game_list = g_list_prepend (game_list, game);
      game -> game_state = scm_make_vector (SCM_MAKINUM (3), SCM_BOOL_F);
      scm_protect_object (game->game_state);
    }

  return SCM_MAKINUM (game -> id);

 error:
  SCM_ASSERT (0, s_rest, SCM_ARGn, s_glue_pool_rules);
  return SCM_UNSPECIFIED;
}

#undef SET_PROCEDURE_ALLOW_F
#undef SET_PROCEDURE

SCM_PROC (s_glue_game_set_event_mask, "game-set-event-mask!", 0,0,1, sglue_game_set_event_mask);

static SCM
sglue_game_set_event_mask (SCM data)
{
  Game *g;
  Animation *a;

  g = gtulpas -> game;
  a = g -> animation;

  glue_parse_event_mask (g, data, SCM_ARG1, s_glue_game_set_event_mask);

  a->ball_ball_num = g -> anim_event_ball_ball_num;
  a->ball_mantinel_num = g -> anim_event_ball_mantinel_num;

  return SCM_UNDEFINED;
}

/*
 *
 *		game & animation events
 */

static void
glue_answers_to_list (Game *game, SCM *last)
{
  GList		*lquery;
  Query		*query;

  SCM		 response;

  *last = SCM_EOL;
  lquery = game -> query_list;

  while (lquery != NULL)
    {
      query = (Query *) (lquery -> data);
      
      switch (query -> response)
	{
	case QUERY_DO_CALL:
	  g_assert_not_reached ();
	  
	case QUERY_DO_IGNORE:
	  response = SCM_BOOL_F;
	  break;
	default:
	  response = SCM_MAKINUM (query->response);
	  break;
	}

      *last = scm_cons (response, SCM_EOL);
      last = SCM_CDRLOC (*last);

      lquery = lquery -> next;
    }
}

static glue_ugly_anim_info*
glue_create_ugly_anim_info (Animation *anim)
{
  static glue_ugly_anim_info info = { SCM_BOOL_F, NULL, NULL, NULL, NULL, NULL, NULL};

  SCM		tmp;

  if (anim -> ugly_ptr != NULL)
    return (glue_ugly_anim_info *) (anim -> ugly_ptr);

  if (info.arg_list != SCM_BOOL_F)
    scm_unprotect_object (info.arg_list);

  tmp = scm_cons (SCM_BOOL_F, SCM_EOL);

  info.answers = SCM_CDRLOC (tmp);
  glue_answers_to_list (anim -> game, info.answers);
  info.game_state = SCM_CARLOC (tmp);

  tmp = scm_cons (SCM_BOOL_F, tmp);
  info.state = SCM_CARLOC (tmp);

  tmp = scm_cons (SCM_BOOL_F, tmp);
  info.info2 = SCM_CARLOC (tmp);

  tmp = scm_cons (SCM_BOOL_F, tmp);
  info.info1 = SCM_CARLOC (tmp);

  tmp = scm_cons (SCM_BOOL_F, tmp);
  info.event_type = SCM_CARLOC (tmp);

  info.arg_list = tmp;

  scm_protect_object (info.arg_list);
  anim -> ugly_ptr = &info;

  return &info;
}

static Animation *glue_process_anim  = NULL;
static guint     glue_process_event_time = 0;

void 
glue_process_anim_event (Animation *anim, AnimEvent *event)
{
  glue_ugly_anim_info *info;

  if ((event->type & anim->game->anim_event_mask) == 0)
    return;

  if (glue_process_anim != NULL)
    return;

  if (event -> type == S_Ball_Ball_Collision &&
      anim -> ball_ball_num == 0)
    return;

  if (event -> type == S_Ball_Mantinel_Collision &&
      anim -> ball_mantinel_num == 0)
    return;

  info = glue_create_ugly_anim_info (anim);

  glue_process_event_time = event -> time_since_bang;
  
  *(info->event_type) = glue_enum_value2scm (event->type, &anim_event_info);
  *(info->game_state) = GAME_STATE (anim -> game);

  switch (event->type)
    {
    case S_Ball_Ball_Collision:
      anim -> ball_ball_num --;
      /* no break */
    case S_Cue_Ball_Collision:
    case S_Ball_Pocketed:
      *(info->info1) = SCM_MAKINUM (event->info1);
      *(info->info2) = SCM_MAKINUM (event->info2);
      break;

    case S_Ball_Mantinel_Collision:
      anim -> ball_mantinel_num --;
      /* no break */
    case S_Ball_Jump_Off:
      *(info->info1) = SCM_MAKINUM (event->info1);
      *(info->info2) = SCM_BOOL_F;
      break;

    case S_End_Of_Shoot:
      *(info->info1) = SCM_BOOL_F;
      *(info->info2) = SCM_BOOL_F;
      break;

    case S_Game_Point:
    case S_Game_Penalty:
    case S_Game_Pocketed: 
    case S_Game_Faul:
    case S_Game_Reward:
      g_assert_not_reached ();
    }

  glue_process_anim = anim;
  *(info->state) = glue_safe_apply (anim->game->game_rules_func, info->arg_list);
  glue_process_anim = NULL;
}

static void
glue_inside_event (const char *proc)
{
  if (glue_process_anim == NULL)
    scm_misc_error (proc,
		    "You can not call this function outside process-event function", 
		    SCM_EOL);
}


SCM_PROC (s_glue_game_reset_score, "game-reset-score", 0,2,0, sglue_game_reset_score);

static SCM
sglue_game_reset_score (SCM s_pid, SCM s_counter)
{
  int		pid;
  int		counter;
  Game		*g;

  g = gtulpas->game;
  SCM_ASSERT (s_pid == SCM_UNDEFINED || SCM_INUMP (s_pid) || s_pid == SCM_BOOL_T, s_pid,
	      SCM_ARG1, s_glue_game_reset_score);
  SCM_ASSERT (s_counter == SCM_UNDEFINED || SCM_INUMP (s_counter), s_counter,
	      SCM_ARG2, s_glue_game_reset_score);

  pid = (s_pid == SCM_BOOL_T) ? -1 : ((s_pid == SCM_UNDEFINED) ? g->cur_player : SCM_INUM (s_pid));
  counter = s_counter == SCM_UNDEFINED ? 0 : SCM_INUM (s_counter);
  
  SCM_ASSERT ((-1 <= pid) && (pid < g -> nof_players), s_pid, SCM_ARG1, s_glue_game_reset_score);
  SCM_ASSERT ((0 <= counter) && (counter < g -> nscore), s_counter, SCM_ARG2, s_glue_game_reset_score);

  game_add_event_nop_type_data (g, RESET_SCORE, (gpointer) event_reset_score_new (counter, pid));

  return SCM_UNSPECIFIED;
}


SCM_PROC (s_glue_game_add_point, "game-add-point", 0,2,0, sglue_game_add_point);
static SCM
sglue_game_add_point (SCM s_counter, SCM s_points)
{
  int points;
  int counter;

  SCM_ASSERT (s_counter == SCM_UNDEFINED || SCM_INUMP (s_counter), s_counter,
	      SCM_ARG1, s_glue_game_add_point);
  SCM_ASSERT (s_points == SCM_UNDEFINED || SCM_INUMP (s_points), s_points,
	      SCM_ARG2, s_glue_game_add_point);

  points  = s_points  == SCM_UNDEFINED ? 1 : SCM_INUM (s_points);
  counter = s_counter == SCM_UNDEFINED ? 0 : SCM_INUM (s_counter);

  if (glue_process_anim)
    animation_add_event (glue_process_anim, glue_process_event_time, 
			 S_Game_Point, counter, points);
  else
    game_add_event_nop_type_data (gtulpas ->game, POINT, (gpointer) event_score_new (counter, points));

  return SCM_UNSPECIFIED;
}

SCM_PROC (s_glue_game_add_penalty, "game-add-penalty", 0,2,0, sglue_game_add_penalty);
static SCM
sglue_game_add_penalty (SCM s_counter, SCM s_points)
{
  int points;
  int counter;

  SCM_ASSERT (s_counter == SCM_UNDEFINED || SCM_INUMP (s_counter), s_counter,
	      SCM_ARG1, s_glue_game_add_penalty);
  SCM_ASSERT (s_points == SCM_UNDEFINED || SCM_INUMP (s_points), s_points,
	      SCM_ARG2, s_glue_game_add_penalty);

  points = s_points == SCM_UNDEFINED ? 1 : SCM_INUM (s_points);
  counter = s_counter == SCM_UNDEFINED ? 0 : SCM_INUM (s_counter);

  if (glue_process_anim)
    animation_add_event (glue_process_anim, glue_process_event_time, 
			 S_Game_Penalty, counter, points);
  else
    game_add_event_nop_type_data (gtulpas ->game, PENALTY, (gpointer) event_score_new (counter, points));
    
  return SCM_UNSPECIFIED;
}

SCM_PROC (s_glue_game_ball_potted, "game-ball-potted",2,0,0, sglue_game_ball_potted);
static SCM
sglue_game_ball_potted (SCM s_counter, SCM s_ball)
{
  int ball;
  int counter;

  SCM_ASSERT (s_counter == SCM_UNDEFINED || SCM_INUMP (s_counter), s_counter,
	      SCM_ARG1, s_glue_game_ball_potted);
  SCM_ASSERT (s_ball == SCM_UNDEFINED || SCM_INUMP (s_ball), s_ball,
	      SCM_ARG2, s_glue_game_ball_potted);

  counter = s_counter == SCM_UNDEFINED ? 0 : SCM_INUM (s_counter);
  ball = s_ball == SCM_UNDEFINED ? 0 : SCM_INUM (s_ball);

  if (glue_process_anim)
    animation_add_event (glue_process_anim, glue_process_event_time, 
			 S_Game_Pocketed, counter, ball);
  else
    game_add_event_nop_type_data (gtulpas ->game, POCKETED, (gpointer) event_score_new (counter, ball));

  return SCM_UNSPECIFIED;
}

SCM_PROC (s_glue_game_ball_out, "game-ball-out",1,0,0, sglue_game_ball_out);
static SCM
sglue_game_ball_out (SCM s_ball)
{
  int  id;
  Game *game;

  glue_inside_event (s_glue_game_ball_out);

  id = SCM_INUMP (s_ball) ? SCM_INUM (s_ball) : -1;
  game = gtulpas->game;

  SCM_ASSERT ((0 <= id) && (id < game->balls_nof), s_ball, SCM_ARG1, s_glue_game_ball_out);
  game_add_event_type_data (gtulpas->game, BALL_OUT, (gpointer) id);
  return SCM_UNSPECIFIED;
}

SCM_PROC (s_glue_game_set_ball_on, "game-set-ball-on", 1, 0,0, sglue_game_set_ball_on);

static SCM
sglue_game_set_ball_on (SCM s_bid)
{
  int	bid;

  bid =  (SCM_INUMP (s_bid)) ? SCM_INUM (s_bid) : -1;
  SCM_ASSERT ((0 <= bid) && (bid < gtulpas->game->balls_nof), 
	      s_bid, SCM_ARG1, s_glue_game_set_ball_on);

  game_add_event_nop_type_data (gtulpas->game,  BALL_ON, (gpointer) bid);
  
  if (game_process_events_inside ())
    gnome_tulpas_score_ball_on (gtulpas->score, bid);

  return SCM_UNSPECIFIED;
}

SCM_PROC (s_glue_game_faul, "game-faul",0,0,0, sglue_game_faul);
static SCM
sglue_game_faul (void)
{
  if (glue_process_anim)
    animation_add_event (glue_process_anim, glue_process_event_time, S_Game_Faul, 0,0);
  else
    animation_add_event (gtulpas->game->animation, 100000, S_Game_Faul, 0,0);

  return SCM_UNSPECIFIED;
}

SCM_PROC (s_glue_game_reward, "game-reward", 0,0,0, sglue_game_reward);
static SCM
sglue_game_reward (void)
{
  if (glue_process_anim)
    animation_add_event (glue_process_anim, glue_process_event_time, S_Game_Reward, 0,0);
  else
    animation_add_event (gtulpas->game->animation, 100000, S_Game_Reward, 0,0);

  return SCM_UNSPECIFIED;
}

SCM_PROC (s_glue_get_event_time, "get-event-time", 0,0,0, sglue_get_event_time);
static SCM
sglue_get_event_time (void)
{
  glue_inside_event (s_glue_get_event_time);
  return SCM_MAKINUM (glue_process_event_time);
}

SCM_PROC (s_glue_get_player_points, "get-player-points", 0, 2, 0, sglue_get_player_points);
static SCM
sglue_get_player_points (SCM s_pid, SCM s_counter)
{
  Game		*g;
  int		pid;
  int		counter;

  g = gtulpas -> game;
  SCM_ASSERT (s_pid == SCM_UNDEFINED || SCM_INUMP (s_pid), s_pid,
	      SCM_ARG1, s_glue_get_player_points);
  SCM_ASSERT (s_counter == SCM_UNDEFINED || SCM_INUMP (s_counter), s_counter,
	      SCM_ARG2, s_glue_get_player_points);

  pid = s_pid == SCM_UNDEFINED ? g->cur_player : SCM_INUM (s_pid);
  counter = s_counter == SCM_UNDEFINED ? 0 : SCM_INUM (s_counter);

  SCM_ASSERT ((0 <= pid) && (pid < g -> nof_players), s_pid, SCM_ARG1, s_glue_get_player_points);
  SCM_ASSERT ((0 <= counter) && (counter < g->nscore), s_counter, SCM_ARG2, s_glue_get_player_points);

  return SCM_MAKINUM (game_get_player_points (g, pid, counter));
}

static int glue_inside_eos = FALSE;

static void
glue_check_inside_eos (const char *proc)
{
  if (glue_inside_eos == FALSE)
    scm_misc_error (proc,
		    "You can not call this function outside end-of-shot function", 
		    SCM_EOL);
}

void
glue_call_end_of_shot (Game *game)
{
  glue_ugly_anim_info *info;
  SCM		balls_out_list;
  SCM		*lptr;
  SCM		asnwers;

  GList		*blist;

  if (scm_procedure_p (game -> end_of_shot_func) == SCM_BOOL_F)
    return;

  glue_inside_eos = TRUE;

  balls_out_list = SCM_EOL;
  lptr = &balls_out_list;

  for (blist = game -> ball_out; blist ; blist = blist -> next)
    {
      *lptr = scm_cons (SCM_MAKINUM ((int) (blist -> data)), SCM_EOL);
      lptr = SCM_CDRLOC (*lptr);
    }
  
  info = glue_create_ugly_anim_info (game->animation);
  glue_safe_apply (game -> end_of_shot_func,
		   scm_cons2 (*(info -> state),
			      GAME_STATE (game),
			      scm_cons (balls_out_list,
					*(info -> answers))));
					
  glue_inside_eos = FALSE;
}

SCM_PROC (s_glue_game_win, "game-win",0,0,0, sglue_game_win);
static SCM
sglue_game_win (void)
{
  glue_check_inside_eos (s_glue_game_win);
  game_add_event_nop_type (gtulpas->game, END_WIN);
  return SCM_UNSPECIFIED;
}


SCM_PROC (s_glue_game_lose, "game-lose",0,0,0, sglue_game_lose);
static SCM
sglue_game_lose (void)
{
  glue_check_inside_eos (s_glue_game_lose);
  game_add_event_nop_type (gtulpas->game, END_LOSE);
  return SCM_UNSPECIFIED;
}

SCM_PROC (s_glue_game_higher_score_wins, "game-higher-score-wins", 0,0,0, sglue_game_higher_score_wins);
static SCM
sglue_game_higher_score_wins (void)
{
  glue_check_inside_eos (s_glue_game_higher_score_wins);
  game_add_event_nop_type (gtulpas->game, END_HIGHER_SCORE);
  return SCM_UNSPECIFIED;
}

SCM_PROC (s_glue_game_next_player, "game-next-player",0,0,0, sglue_game_next_player);
static SCM
sglue_game_next_player (void)
{
  glue_check_inside_eos (s_glue_game_next_player);
  game_add_event_nop_type (gtulpas->game, NEXT_PLAYER);
  return SCM_UNSPECIFIED;
}

SCM_PROC (s_glue_game_set_player, "game-set-player",1,0,0, sglue_game_set_player);
static SCM
sglue_game_set_player (SCM s_pid)
{
  int	pid = -1;

  glue_check_inside_eos (s_glue_game_set_player);
  if (SCM_INUMP (s_pid))
    pid = SCM_INUM (s_pid);

  SCM_ASSERT ((0 <= pid) && (pid < gtulpas -> game -> nof_players), s_pid,
	      SCM_ARG1, s_glue_game_set_player);

  game_add_event_nop_type_data (gtulpas->game, SET_PLAYER, (gpointer) pid);
  return SCM_UNSPECIFIED;
}


/*
 *	game state
 */

SCM_PROC (s_glue_game_state, "game-state", 0, 0, 0, sglue_game_state);
static SCM
sglue_game_state (void)
{
  /* usually useful only for console - ... */
  return GAME_STATE (gtulpas->game);
}

SCM_PROC (s_glue_game_state_set_x, "game-state-set!", 1, 0, 0, sglue_game_event_set_x);

static SCM
sglue_game_event_set_x (SCM new_state)
{
  Game	*game;

  /* KILLS UNDO -+> so we disable undo !!! */
  gtulpas_undo_sensitivity (FALSE);

  game = gtulpas->game;
  GAME_STATE (game) = new_state;

  return SCM_UNSPECIFIED;
}

void
glue_game_state_backup (Game *game)
{
  gtulpas_undo_sensitivity (TRUE);

  GAME_STATE_BUP_2 (game) = GAME_STATE_BUP_1 (game);
  GAME_STATE_BUP_1 (game) = 
      (scm_procedure_p (game->game_state_deep_copy_func) == SCM_BOOL_F)
    ? GAME_STATE (game) 
    : glue_safe_call1 (game->game_state_deep_copy_func,
		       GAME_STATE (game));
}

void
glue_game_state_restore (Game *game)
{
  GAME_STATE (game) = GAME_STATE_BUP_1 (game);
  GAME_STATE_BUP_1 (game) = GAME_STATE_BUP_2 (game);
}

/*
 *	balls coords
 */ 

SCM_PROC (s_glue_ball_get_x, "ball-get-x", 1, 1, 0, sglue_ball_get_x);

static SCM
sglue_ball_get_x (SCM s_id, SCM s_time)
{
  int		id;
  int		time;
  Game		*g;
  Animation	*a;

  g = gtulpas -> game;
  a = g->animation;

  id = SCM_INUMP (s_id) ? SCM_INUM (s_id) : -1;

  if (s_time == SCM_UNDEFINED || s_time == SCM_BOOL_F)
    time = a -> length - 1;
  else if (SCM_INUMP (s_time))
    time = (int) floor (((dbl) SCM_INUM (time)) / 1000.0 * POOL_DEFAULT_BIG_MOVER_FPS);
  else
    time = -1;

  SCM_ASSERT ((0 <= id) && (id < g->balls_nof), s_id, SCM_ARG1, s_glue_ball_get_x);
  SCM_ASSERT ((0 <= time) && (time < a->length) , s_time, SCM_ARG1, s_glue_ball_get_x);

  return gh_double2scm ((a->balls + a->nof_balls * time + id) -> p[X]);
}

SCM_PROC (s_glue_ball_get_y, "ball-get-y", 2, 0, 0, sglue_ball_get_y);

static SCM
sglue_ball_get_y (SCM s_id, SCM s_time)
{
  int		id;
  int		time;
  Game		*g;
  Animation	*a;

  g = gtulpas -> game;
  a = g->animation;

  id = SCM_INUMP (s_id) ? SCM_INUM (s_id) : -1;
  time = SCM_INUMP (s_time) ? SCM_INUM (s_time) : -1;

  if (s_time == SCM_UNDEFINED || s_time == SCM_BOOL_F)
    time = a -> length - 1;
  else if (SCM_INUMP (s_time))
    time = (int) floor (((dbl) SCM_INUM (time)) / 1000.0 * POOL_DEFAULT_BIG_MOVER_FPS);
  else
    time = -1;

  SCM_ASSERT ((0 <= id) && (id < g->balls_nof), s_id, SCM_ARG1, s_glue_ball_get_y);
  SCM_ASSERT ((0 <= time) && (time < a->length) , s_time, SCM_ARG1, s_glue_ball_get_y);

  return gh_double2scm ((a->balls + a->nof_balls * time + id) -> p[Y]);
}


/*
 *
 *	enum -=> symbol translation
 */

static int
glue_enum_item_cmp_by_symbol (const void *p1, const void *p2)
{
  return (((glue_enum_item *) p1) -> symbol) - (((glue_enum_item *) p2) -> symbol);
}

static int
glue_enum_item_cmp_by_value (const void *p1, const void *p2)
{
  return (((glue_enum_item *) p1) -> value) - (((glue_enum_item *) p2) -> value);
}

static int
glue_enum_item_key_symbol (const void *p1, const void *p2)
{
  return  (* ((SCM *) p1)) - (((glue_enum_item *) p2) -> symbol);
}

static int
glue_enum_item_key_value (const void *p1, const void *p2)
{
  return (* ((int *) p1)) - (((glue_enum_item *) p2) -> value);
}

void
glue_enum_info_init (glue_enum_info *info)
{
  glue_enum_item	*items;
  int			 i, num;

  items = info -> sorted_by_symbol;
  num = info -> num_of_items;

  for (i = 0; i < info->num_of_items; i++)
    scm_protect_object (items [i].symbol = SCM_CAR (scm_intern0 (items [i].name)));

  memcpy (info->sorted_by_value, items, num * sizeof (glue_enum_item));

  qsort (info->sorted_by_symbol, num, sizeof (glue_enum_item), glue_enum_item_cmp_by_symbol);
  qsort (info->sorted_by_value,  num, sizeof (glue_enum_item), glue_enum_item_cmp_by_value);
}

SCM
glue_enum_value2scm (int key_value, glue_enum_info *info)
{
  glue_enum_item *bret;

  bret = (glue_enum_item *)
    bsearch (&key_value, info->sorted_by_value, 
	     info->num_of_items, sizeof (glue_enum_item),
	     glue_enum_item_key_value);
		 
  return (bret == NULL) ? SCM_UNDEFINED : bret -> symbol;
}

int
glue_enum_scm2value (SCM key_symbol, glue_enum_info *info)
{
  glue_enum_item *bret;

  bret = (glue_enum_item *)
    bsearch (&key_symbol, info->sorted_by_symbol,
	     info->num_of_items, sizeof (glue_enum_item),
	     glue_enum_item_key_symbol);

  return (bret == NULL) ? -1 : bret -> value;
}

int
glue_enum_scmlist2value (SCM list, glue_enum_info *info, int narg, const char *name)
{
  int		rval, cret;

  rval = 0;
  while (list != SCM_EOL)
    {
      if (! SCM_CONSP (list))
	goto error;

      cret = glue_enum_scm2value (SCM_CAR (list), info);

      if (cret == -1)
	goto error;
      
      rval |= cret;

      list = SCM_CDR (list);
    }

  return rval;

 error:
  SCM_ASSERT (0, list, narg, name);
  return -1;
}

/*
 *
 *		other general glue thingies
 */

extern int scm_block_gc;

void
glue_gc_disable (void)
{
#ifdef DEBUG
  g_message ("Guile GC (%d) disabled", scm_block_gc);
#endif
  scm_block_gc = 1;
}

void
glue_gc_enable (void)
{
#ifdef DEBUG
  g_message ("Guile GC (%d) enabled", scm_block_gc);
#endif
  scm_block_gc = 0;
	  /* when needed -=> force GC */
  scm_cons (SCM_EOL, SCM_EOL);
}

SCM_SYMBOL (glue_error_occured, "error-occured");

SCM
glue_catch_handler (void *null_data, SCM tag, SCM throw_args)
{
  SCM		stack;
  SCM		port;
  SCM		use_gtk = 0;

  if (gtk_main_level () > 0)
    {
      use_gtk = 1;
      port = scm_mkstrport (SCM_INUM0,
			    scm_make_string (SCM_INUM0, SCM_UNDEFINED),
			    SCM_OPN | SCM_WRTNG,
			    "glue-catch-handler");
    }
  else
    port = scm_current_output_port ();

  stack = scm_fluid_ref (SCM_CDR (scm_the_last_stack_fluid));

  scm_display_backtrace (stack, port, SCM_UNDEFINED, SCM_UNDEFINED);
  scm_display_error (stack,port,
		     SCM_CAR (throw_args),
		     SCM_CADR (throw_args),
		     SCM_CADDR (throw_args),
		     SCM_CDDDR (throw_args));

  
  if (use_gtk)
    {
      SCM	errstr;

      errstr = scm_strport_to_string (port);

      fwrite (SCM_CHARS (errstr), 1, SCM_LENGTH (errstr), stderr);
      gtulpas_glue_error (SCM_CHARS (errstr), SCM_LENGTH (errstr));
    }

  return glue_error_occured;
}

static struct {
  SCM	function;
  SCM	args;
} glue_catch_apply_data;

static SCM
glue_catch_apply_dispatcher (void)
{
  return scm_apply (glue_catch_apply_data.function,
		    glue_catch_apply_data.args,
		    SCM_EOL);
}

SCM
glue_safe_apply (SCM function, SCM args)
{
  glue_catch_apply_data.function = function;
  glue_catch_apply_data.args = args;

  return SAFE_SCHEME_CALL (glue_catch_apply_dispatcher, NULL);
}

SCM_PROC (s_glue_safe_call, "safe-call", 2, 0, 0, sglue_safe_call);

SCM
sglue_safe_call (SCM s_function, SCM s_recover)
{
  SCM  retval;

  SCM_ASSERT (scm_procedure_p (s_function) == SCM_BOOL_T, s_function, SCM_ARG1, s_glue_safe_call);
  SCM_ASSERT (scm_procedure_p (s_recover) == SCM_BOOL_T, s_recover, SCM_ARG2, s_glue_safe_call);

  retval = glue_safe_call0 (s_function);

  if (retval == glue_error_occured)
    return glue_safe_call0 (s_recover);

  return retval;
}

static struct
{
  int arguments;

  SCM function;
  SCM a1;
  SCM a2;
  SCM a3;
  SCM a4;
} glue_catch_call_data;

static SCM
glue_catch_call_dispatcher (void)
{
  switch (glue_catch_call_data.arguments)
    {
    case 0:
      return gh_call0 (glue_catch_call_data.function);
    case 1:
      return gh_call1 (glue_catch_call_data.function,
		       glue_catch_call_data.a1);
    case 2:
      return gh_call2 (glue_catch_call_data.function,
		       glue_catch_call_data.a1,
		       glue_catch_call_data.a2);
    case 3:
      return gh_call3 (glue_catch_call_data.function,
		       glue_catch_call_data.a1,
		       glue_catch_call_data.a2, 
		       glue_catch_call_data.a3); 
    case 4:
      return scm_apply (glue_catch_call_data.function,
			scm_cons2 (glue_catch_call_data.a1,
				   glue_catch_call_data.a2, 
				   scm_cons2 (glue_catch_call_data.a3,
					      glue_catch_call_data.a4,
					      SCM_EOL)),
			SCM_EOL);
    default:
      g_assert_not_reached ();
    }
  return SCM_UNSPECIFIED;
}

SCM
glue_safe_call0 (SCM function)
{
  glue_catch_call_data.arguments = 0;
  glue_catch_call_data.function = function;

  return SAFE_SCHEME_CALL (glue_catch_call_dispatcher, NULL);
}

SCM
glue_safe_call1 (SCM function, SCM arg1)
{
  glue_catch_call_data.arguments = 1;
  glue_catch_call_data.function = function;
  glue_catch_call_data.a1 = arg1;

  return SAFE_SCHEME_CALL (glue_catch_call_dispatcher, NULL);
}

SCM
glue_safe_call2 (SCM function, SCM arg1, SCM arg2)
{
  glue_catch_call_data.arguments = 2;
  glue_catch_call_data.function = function;
  glue_catch_call_data.a1 = arg1;
  glue_catch_call_data.a2 = arg2;

  return SAFE_SCHEME_CALL (glue_catch_call_dispatcher, NULL);
}

SCM
glue_safe_call3 (SCM function, SCM arg1, SCM arg2, SCM arg3)
{
  glue_catch_call_data.arguments = 3;
  glue_catch_call_data.function = function;
  glue_catch_call_data.a1 = arg1;
  glue_catch_call_data.a2 = arg2;
  glue_catch_call_data.a3 = arg3;

  return SAFE_SCHEME_CALL (glue_catch_call_dispatcher, NULL);
}

SCM
glue_safe_call4 (SCM function, SCM arg1, SCM arg2, SCM arg3, SCM arg4)
{
  glue_catch_call_data.arguments = 4;
  glue_catch_call_data.function = function;
  glue_catch_call_data.a1 = arg1;
  glue_catch_call_data.a2 = arg2;
  glue_catch_call_data.a3 = arg3;
  glue_catch_call_data.a4 = arg4;

  return SAFE_SCHEME_CALL (glue_catch_call_dispatcher, NULL);
}

void
glue_debug_console (void)
{
#ifdef DEBUG
  SAFE_SCHEME_CALL (gh_eval_str, "(debug-console)");
#endif
}

SCM_VCELL (sglue_data_dir, "tulpas-data-dir");
SCM_VCELL (sglue_developer, "tulpas-developers-mode");

static char *glue_init_file_name = NULL;

void
glue_load_scheme_files (void)
{
  SAFE_SCHEME_CALL (gh_eval_file, glue_init_file_name);
}

void
gtulpas_guile_init (void)
{
  Skin		*skin;
  FyjikSkin	*fskin;
  char		*path;

#ifndef SCM_MAGIC_SNARFER
#include "guileglue.x"
#endif

#define SET_COLOR(X,R,G,B)   X[CR] = R; X[CG]=G; X[CB]=B

  glue_enum_info_init (&anim_event_info);
  glue_enum_info_init (&glue_pool_rules_info);
  glue_enum_info_init (&glue_score_types_info);

  default_fyjik_skin = fskin = g_new (FyjikSkin, 1);
  *fskin = builtin_fyjik_skin;
  fskin -> name = g_strdup ("default-fyjik-skin");

  /* default ball skin */

  default_ball_skin = skin = g_new (Skin, 1);
  skin -> name = g_strdup ("default-ball-skin");

  SET_COLOR (skin -> color1, 1,1,1);
  SET_COLOR (skin -> color2, 0,0,0);

  skin -> texture_file = NULL;
  skin -> dvalue = 0;
  skin -> has_strip = 0;
  skin -> has_mark = 0;

  skin_list = g_list_append (skin_list, skin);

  /* cue skin */

  pool_cue.cue_skin = default_cue_skin = skin = g_new (Skin, 1);
  skin -> name = g_strdup ("default-cue-skin");

  SET_COLOR (skin -> color1, 255. / 255., 165. / 255., 79. / 255.);
  SET_COLOR (skin -> color2, 0.32, 0.66, 0.88);

  skin -> texture_file = NULL;
  skin -> dvalue = 1;
  skin -> has_strip = 0;
  skin -> has_mark = 0;
  
  skin_list = g_list_append (skin_list, skin);

  /* mark skin */

  default_mark_skin = skin = g_new (Skin, 1);
  *skin = *default_ball_skin;

  skin -> name = g_strdup ("default-mark-skin");
  SET_COLOR (skin->color1, /* gainsboro */ 220./255., 220./255., 220./255.);

  skin_list = g_list_append (skin_list, skin);

  /* table skin */

  default_table_skin = skin = g_new (Skin, 1);
  *skin = *default_ball_skin;

  skin -> name = g_strdup ("default-table-skin");
  SET_COLOR (skin->color1, /* SeaGreen */ 46./255., 139./255., 87./255.);
  skin_list = g_list_append (skin_list, skin);    

  /* table pocket skin */

  default_pocket_skin = skin = g_new (Skin, 1);
  *skin = *default_ball_skin;

  skin -> name = g_strdup ("default-pocket-skin");
  SET_COLOR (skin->color1, 0.6, 0.6, 0.6);
  skin_list = g_list_append (skin_list, skin);    

  /* table outline skin */

  pool_table.outer_skin = default_outline_skin = skin = g_new (Skin, 1);
  *skin = *default_ball_skin;

  skin -> name = g_strdup ("default-table-outline-skin");
  SET_COLOR (skin->color1, 165./255.,  42./255., 42./255.);
  skin_list = g_list_append (skin_list, skin);    

#ifdef DEBUG
  gh_eval_str ("(debug-enable 'debug)");

  SCM_SETCDR (sglue_data_dir, gh_str02scm (".."));
  SCM_SETCDR (sglue_developer, SCM_BOOL_T);

  glue_init_file_name = g_strdup ("../scheme/init.scm");
#else
  path = gnome_unconditional_datadir_file ("gtulpas");

  SCM_SETCDR (sglue_data_dir, gh_str02scm (path));
  g_free (path);

  glue_init_file_name = gnome_unconditional_datadir_file ("gtulpas/scheme/init.scm");
#endif

  glue_load_scheme_files ();

  if ((game_list == NULL) || (pool_cue_type_list == NULL))
    {
      g_error ("No rules defined during load of init.scm :(");
      exit (1);
    }
}
