#include <GL/gl.h>
#include <math.h>

#include "mathutils.h"

#ifndef PI
#define PI 3.14159265358979323846264338328 
#endif

void vec_cross (dbl *r, dbl *u, dbl *v)
{
  if (r == u || r == v)
    {
      dbl	rr[3];

      rr[0] = u[1] * v[2] - u[2] * v[1];
      rr[1] = u[2] * v[0] - u[0] * v[2];
      rr[2] = u[0] * v[1] - u[1] * v[0];

      r[0] = rr[0];
      r[1] = rr[1];
      r[2] = rr[2];
    }
  else
    {
      r[0] = u[1] * v[2] - u[2] * v[1];
      r[1] = u[2] * v[0] - u[0] * v[2];
      r[2] = u[0] * v[1] - u[1] * v[0];
    }
}

void vec_norm (dbl *result, dbl *u)
{
  dbl l;

  l = sqrt (u[0]*u[0] + u[1]*u[1] + u[2]*u[2]);

  if (l < EPSILON)
    return;

  /* l = 1.0 / l; */
  result[0] = u[0] / l;
  result[1] = u[1] / l;
  result[2] = u[2] / l;
}

void vec_to_sph(dbl *result, dbl *vec)
{
  double x= vec[X],
    y= vec[Z],
    z= vec[Y],

    r;

  r= x*x+y*y+z*z;
  
  if (r==0.0)
    { result[0]=result[1]=result[2]=0.0; return; }

  r= sqrt(r);
  
  x /= r;
  y /= r;
  z /= r;

  result[0]= r;

  result[1]= asin(y);   /* phi */
  r= x*x+z*z;

  if (r == 0)
	  result[2]=0;	/* theta */
  else
    {
      if (z==0.0)
	if (x > 0.0)
	  result[2]= 0;
	else
	  result[2]= PI;
      else
	{
	  result[2]= acos(x/sqrt(r));
	  if (z < 0.0)
	    result[2] = 2*PI - result[2];
	}
    }
}


void sph_to_vec(dbl *result, dbl *vec)
{
  double tmp, vec2;

  tmp= vec[0] * cos(vec[1]);

  vec2 = vec[2];	/* we should safe v[2] since result could be same as vec */

  result[2]= vec[0] * sin(vec[1]);
  result[0]= tmp * cos(vec2);
  result[1]= tmp * sin(vec2);
}

void pool_vector_rotate (dbl *vector,
			 dbl h_change,
			 dbl v_change)
{
 vec_to_sph (vector, vector);

  vector [1]+= v_change;
  vector [2]+= h_change;

  if (vector [1] < (-PI/2 - EPSILON))
    vector [1]= - PI/2 - EPSILON;

  if (vector [1] > PI/2 + EPSILON)
    vector [1]= PI/2 + EPSILON;

  sph_to_vec(vector ,vector);
}

void 
angles_to_vectors (dbl alfa, dbl beta, dbl *orientation, dbl *right, dbl *up)
{
  dbl		 tmp;

  orientation [X] = - cos (alfa);  
  orientation [Y] = - sin (alfa);
  orientation [Z] = 0;

  right [X] = orientation [Y];
  right [Y] = - orientation [X];
  right [Z] = 0.0;

  tmp = cos (beta);

  orientation [X] *= tmp;
  orientation [Y] *= tmp;
  orientation [Z] = - sin (beta);

  vec_cross (up, right, orientation);
}


int solve_quadr_eq( dbl a, dbl b, dbl c, dbl *t1, dbl *t2 )
{
  dbl 	d;

  if ( fabs( a ) < EPSILON )
    {
      if ( fabs( b ) < EPSILON )
	return 0;
      else
	{
	  *t1 = *t2 = - c / b;
	  return 1;
	}
    }

  d = sqr(b) - 4*a*c;

  if ( d < 0.0 )
    return 0;

  d = sqrt( d );

  *t1 = -b + d;
  *t2 = -b - d;

  *t1 /= 2*a;
  *t2 /= 2*a;

  return 1;
}

void get_point( dbl *dst, dbl *s, dbl *dir, dbl t )
{
  dst[0] = dir[0]*t + s[0];
  dst[1] = dir[1]*t + s[1];
  dst[2] = dir[2]*t + s[2];
}

/*
  from www.gamasutra.com:

  http://www.gamasutra.com/features/programming/19980703/quaternions_07.htm

  http://www.gamasutra.com/features/programming/19980703/quaternions_08.htm

*/
/*
  http://www.cs.berkeley.edu/~laura/cs184/quat/quaternion.html - actual used equation 
*/

#define w	0
#define x	1
#define y	2
#define z	3

void quat_mul( dbl *r, dbl *q1, dbl *q2 )
{
  dbl	tmp[4], *res;
  dbl   t1, t2, t3;
  int	b;

  if ( (b = (r == q1 || r == q2)) )
    res = tmp;
  else
    res = r;

  t1 = q1[0];
  t2 = q2[0];
  t3 = t1 * t2;
  printf ("%g %g\n", t1 * t2, t3);


  res[w] = q1[w]*q2[w] - q1[x]*q2[x] - q1[y]*q2[y] - q1[z]*q2[z];
  res[x] = q1[w]*q2[x] + q1[x]*q2[w] + q1[y]*q2[z] - q1[z]*q2[y];
  res[y] = q1[w]*q2[y] - q1[x]*q2[z] + q1[y]*q2[w] + q1[z]*q2[x];
  res[z] = q1[w]*q2[z] + q1[x]*q2[y] - q1[y]*q2[x] + q1[z]*q2[w];

  if ( b )
    {
      r[0] = res[0];
      r[1] = res[1];
      r[2] = res[2];
      r[3] = res[3];
    }
}

#undef w 
#undef x 
#undef y 
#undef z 

#define w	q[0]
#define x	q[1]
#define y	q[2]
#define z	q[3]

void quat_to_rmat( dbl m[4][4], dbl *q )
{
  m[0][0] = w*w + x*x - y*y -z*z;
  m[0][1] = 2*x*y + 2*w*z;
  m[0][2] = 2*x*z - 2*w*y;
  m[0][3] = 0.0;

  m[1][0] = 2*x*y - 2*w*z;
  m[1][1] = w*w - x*x + y*y - z*z;
  m[1][2] = 2*y*z + 2*w*x;
  m[1][3] = 0.0;

  m[2][0] = 2*x*z + 2*w*y;
  m[2][1] = 2*y*z - 2*w*x;
  m[2][2] = w*w - x*x - y*y + z*z;
  m[2][3] = 0.0;

  m[3][0] = 0.0;
  m[3][1] = 0.0;
  m[3][2] = 0.0;
  m[3][3] = 1.0;
}

/* v must be unit vector */
void quat_aaxis( dbl *q, dbl theta, dbl *v )
{
  q[0] = cos(theta/2);	
  
  vec_mul_c( q + 1, v, sin(theta/2) );
}

void quat_norm( dbl *qd, dbl *q )
{
  dbl	norm;

  norm = sqrt( w*w + x*x + y*y + z*z );

  if ( norm == 0.0 )
    return;

  qd[0] = w / norm;
  qd[1] = x / norm;
  qd[2] = y / norm;
  qd[3] = z / norm;
}

void quat_load_from_angles (dbl *q, dbl around_x, dbl around_z)
{
  static dbl ax [] = {1,0,0};
  static dbl az [] = {0,0,1};

  dbl   q1 [4], q2 [4];

  quat_aaxis (q1, around_x, ax);
  quat_aaxis (q2, around_z, az);

  quat_mul (q, q2, q1);
  quat_norm (q,q);
}

#undef w 
#undef x 
#undef y 
#undef z 

void mat_to_mat( dbl dst[4][4], dbl src[4][4] )
{
  int 	i, j;

  for( i = 0; i < 4; i++ )
    for( j = 0; j < 4; j++ )
      dst[i][j] = src[j][i];
}


