#ifndef __POOL_MATH_UTILS_H__
#define __POOL_MATH_UTILS_H__

typedef double dbl;

#ifndef PI
#define PI  3.14159265358979323846264338328
#endif

#define EPSILON				10e-4
#define SLIGHTLY_ABOVE_TABLE		0.1

#ifndef MAX
#define MAX(X,Y)			((X)>(Y)?(X):(Y))
#endif

#ifndef MIN
#define MIN(X,Y)			((X)>(Y)?(Y):(X))
#endif

/*
 *
 *	Vector operations...
 */

#define vec_cpy(DEST, SRC)		do { (DEST)[X] = (SRC)[X];		\
					     (DEST)[Y] = (SRC)[Y];		\
					     (DEST)[Z] = (SRC)[Z];		\
					} while (0)

#define vec_mul_c(RESULT, U, C)		do {  (RESULT)[0] = (C) * (U)[0];	\
					      (RESULT)[1] = (C) * (U)[1];	\
					      (RESULT)[2] = (C) * (U)[2];	\
					} while (0)

#define vec_dot(V1, V2)			((V1)[0]*(V2)[0] + (V1)[1]*(V2)[1] + (V1)[2]*(V2)[2])
#define vec_size2(V)			((V)[0]*(V)[0] + (V)[1]*(V)[1] + (V)[2]*(V)[2])
#define vec_size(V) 			sqrt(vec_size2(V))
#define vec_neg(V) 			do {(V)[0] = -(V)[0]; \
                                            (V)[1] = -(V)[1]; \
					    (V)[2] = -(V)[2]; } while (0)

#define vec_add(R,A,B)			do { (R)[0] = (A)[0] + (B)[0];  \
					     (R)[1] = (A)[1] + (B)[1];  \
					     (R)[2] = (A)[2] + (B)[2]; } while (0)

#define vec_sub(R,A,B)			do { (R)[0] = (A)[0] - (B)[0];  \
					     (R)[1] = (A)[1] - (B)[1];  \
					     (R)[2] = (A)[2] - (B)[2]; } while (0)


#define compPlaneCoefD(N, V) 		(- ( (N)[0]*(V)[0] + (N)[1]*(V)[1] + (N)[2]*(V)[2] ))

#define cvec(D, S)			vec_cpy((D),(S))

#define sqr( X )			((X)*(X))

#define euclid( v1, v2 ) 		sqrt(  sqr( (v1)[0] - (v2)[0] ) + \
					       sqr( (v1)[1] - (v2)[1] ) + \
					       sqr( (v1)[2] - (v2)[2] ) ) 

#define euclid2( v1, v2 ) 		(  sqr( (v1)[0] - (v2)[0] ) + \
					   sqr( (v1)[1] - (v2)[1] ) + \
					   sqr( (v1)[2] - (v2)[2] ) ) 

#define vec_scl( n, v ) 		vec_mul_c ((v), (v), (n))

#define vec_angle( V1, V2 ) 		((vec_dot( V1, V2 ) < ( vec_size( V1 ) * vec_size( V2 ))) ? acos( vec_dot( V1, V2 ) / ( vec_size( V1 ) * vec_size( V2 ) ) ) : 0)

#define RAD2DEG(A)			((A) * 180.0 / PI)
#define DEG2RAD(A)			((A) * PI / 180.0)

#define det2( A11, A12, A21, A22 )	( (A11)*(A22) - (A12)*(A21) )

#define quat_load_identity(Q)		do { (Q)[0] = 1.0;(Q)[1] = 0;(Q)[2] = 0;(Q)[3] = 0; } while (0)
     /*
      *	 usefull constants ...
      */

#define X			0
#define Y			1
#define Z			2
#define W			3

#define CR			0
#define CG			1
#define CB			2
#define CA			3

void vec_cross (dbl *r, dbl *u, dbl *v);
void vec_norm (dbl *result, dbl *u);

void vec_to_sph(dbl *result, dbl *vec);
void sph_to_vec(dbl *result, dbl *vec);

int  solve_quadr_eq( dbl a, dbl b, dbl c, dbl *t1, dbl *t2 );
void get_point( dbl *dst, dbl *s, dbl *dir, dbl t );

void pool_vector_rotate (dbl *vector, dbl h_change, dbl v_change);
void angles_to_vectors (dbl alpha, dbl beta, dbl *orientation, dbl *right, dbl *up);

void quat_mul( dbl *res, dbl *q1, dbl *q2 );
void quat_to_rmat( dbl m[4][4], dbl *quat );
void quat_aaxis( dbl *q, dbl theta, dbl *v );
void quat_norm( dbl *qd, dbl *q );
void quat_load_from_angles (dbl *q, dbl around_x, dbl around_y);

void mat_to_mat( dbl dst[4][4], dbl src[4][4] );

/*
 *
 *	Sphere tesselation
 */

extern dbl 	 *ts_vertices;
extern dbl 	 *ts_normals;
extern int	 *ts_indices;

void	 ts_free (void);
void	 ts_sphere_create (dbl radius, int xy, int z);

void	 ts_sphere_cap (dbl radius, dbl s_width, int xy, int z);
void	 ts_sphere_strip (dbl radius, dbl s_width, int xy, int z);
void	 ts_sphere_mark (dbl radius, dbl m_radius, int xy, int z);


#endif
